/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas.Compiler {

    /**
     * Class for lexical analyzers.
     */
    public class Analyzer : GLib.Object {

        /* The token iterator. */
        private Gee.Iterator<Token> iterator;
        /* Nesting parenthesis level. */
        private int par_level;

        /**
         * Analyzes a list of tokens.
         * @param tokens the list of tokens to analyze.
         * @return an internal representation tree.
         */
        public Tree analyze(Gee.ArrayList<Token> tokens)
        throws SyntaxError {
            iterator = tokens.iterator();
            par_level = 0;
            var root = analyze_expression();
            while (iterator.has_next())
                root = analyze_next_token(root);
            if (par_level != 0)
                throw get_error("Unbalanced parenthesis");
            return root;
        }

        /* Analyzes the next available token. */
        private Tree? analyze_next_token(Tree left) throws SyntaxError {
            if (!iterator.has_next() && par_level > 0)
                throw get_error("Missing closing parenthesis");
            if (!iterator.next())
                return left;
            var token = iterator.get();
            if (token.kind == Token.Kind.BINARY_OPERATOR) {
                Tree tree = new Tree();
                tree.token = token;
                tree.left = left;
                tree.right = analyze_expression();
                return tree;
            }
            if (token.kind == Token.Kind.PAR_RIGHT) {
                par_level--;
                return analyze_next_token(left);
            }
            throw get_error("Unexpected token '%s'".printf(
                                token.original));
        }

        /* Analyzes a whole expression. */
        private Tree? analyze_expression() throws SyntaxError {
            if (!iterator.next())
                throw get_error("Incomplete expresion");
            var token = iterator.get();
            switch (token.kind) {
            case Token.Kind.PAR_LEFT:
                par_level++;
                var expr = analyze_expression();
                return analyze_next_token(expr);
            case Token.Kind.PAR_RIGHT:
                throw get_error("Unexpected right parenthesis");
            case Token.Kind.BINARY_OPERATOR:
                throw get_error("Unexpected operator '%s'".printf(
                                    token.original));
            case Token.Kind.UNARY_OPERATOR:
                var tree = new Tree();
                tree.token = token;
                tree.right = analyze_expression();
                return tree;
            case Token.Kind.FIELD:
                return analyze_match(token);
            case Token.Kind.COLON:
                throw get_error("Unexpected colon");
            case Token.Kind.STRING:
                throw get_error("Unexpected string '%s'".printf(
                                    token.original));
            }
            return null;
        }

        /* Analyzes a match. */
        private Tree? analyze_match(Token field) throws SyntaxError {
            if (!iterator.next())
                throw get_error("Missing colon");
            var colon = iterator.get();
            if (!iterator.next())
                throw get_error("Missing string");
            var str = iterator.get();
            if (colon.kind != Token.Kind.COLON)
                throw get_error("Invalid colon '%s'".printf(colon.original));
            if (str.kind != Token.Kind.STRING)
                throw get_error("Invalid string '%s'".printf(str.original));
            Tree tree = new Tree();
            tree.token = colon;
            tree.left = new Tree();
            tree.left.token = field;
            tree.right = new Tree();
            tree.right.token = str;
            return tree;
        }
    }
}
