/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas.Compiler {

    /**
     * Class for tokenizers.
     */
    public class Tokenizer : GLib.Object {

        /* The tokenizer states. */
        private enum State {
            NORMAL,
            IN_STRING,
            IN_QUOTES;
        }

        /* Title string. */
        private const string TITLE = "title";
        /* Artist string. */
        private const string ARTIST = "artist";
        /* Album string. */
        private const string ALBUM = "album";

        /* The current position on the input string. */
        private int pos;
        /* The previous position on the input string. */
        private int prev_pos;
        /* The input string. */
        private string input;
        /* The list of tokens. */
        private Gee.ArrayList<Token> tokens;
        /* The current state. */
        private State state;
        /* Initial index of the current string. */
        private int ini_str;

        /**
         * Tokenizes an input string.
         * @param input the input string.
         * @return a list of tokens.
         */
        public Gee.ArrayList<Token> tokenize(string input)
        throws SyntaxError {
            prev_pos = pos = 0;
            this.input = input;
            tokens = new Gee.ArrayList<Token>();
            state = State.NORMAL;
            ini_str = 0;

            unichar c;
            while (input.get_next_char(ref pos, out c)) {
                switch (state) {
                case State.NORMAL:
                    next_char_normal(c);
                    break;
                case State.IN_STRING:
                    next_char_string(c);
                    break;
                case State.IN_QUOTES:
                    next_char_quotes(c);
                    break;
                }
                prev_pos = pos;
            }
            var n = input.length;
            switch (state) {
            case State.IN_QUOTES:
                throw get_error("Unclosed quote: %s".printf(
                                    input[ini_str-1:n]));
            case State.IN_STRING:
                tokens.add(get_token());
                break;
            }
            return tokens;
        }

        /* Processes the next character on the normal state. */
        private void next_char_normal(unichar c) throws SyntaxError {
            var str = c.to_string();
            switch (str) {
            case "(":
                tokens.add(new ParLeftToken(pos));
                break;
            case ")":
                tokens.add(new ParRightToken(pos));
                break;
            case ":":
                tokens.add(new ColonToken(pos));
                break;
            case "&":
                tokens.add(new OperatorAndToken(pos));
                break;
            case "|":
                tokens.add(new OperatorOrToken(pos));
                break;
            case "!":
                tokens.add(new OperatorNotToken(pos));
                break;
            case "\"":
                ini_str = pos;
                state = State.IN_QUOTES;
                break;
            default:
                if (c.isspace()) {
                    return;
                } else if (c.isalpha()) {
                    ini_str = prev_pos;
                    state = State.IN_STRING;
                } else {
                    throw get_error("Invalid symbol '%s'".printf(
                                        c.to_string()));
                }
                break;
            }
        }

        /* Processes the next character on the string state. */
        private void next_char_string(unichar c) throws SyntaxError {
            if (c.isalpha())
                return;
            pos--;
            tokens.add(get_token());
            pos++;
            state = State.NORMAL;
            next_char_normal(c);
        }

        /* Processes the next character on the quotes state. */
        private void next_char_quotes(unichar c) {
            var str = c.to_string();
            if (str != "\"")
                return;
            pos--;
            tokens.add(get_token());
            pos++;
            state = State.NORMAL;
        }

        /* Gets a field or string token. */
        private Token get_token() {
            string f = input[ini_str:pos];
            var l = f.casefold();
            if (TITLE.has_prefix(l) && l.length <= TITLE.length)
                return new TitleFieldToken(ini_str);
            if (ARTIST.has_prefix(l) && l.length <= ARTIST.length)
                return new ArtistFieldToken(ini_str);
            if ((ALBUM.has_prefix(l) && l.length <= ALBUM.length) ||
                l == "l")
                return new AlbumFieldToken(ini_str);
            return new StringToken(ini_str, f);
        }
    }
}
