/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas.Compiler {

    /**
     * Class for translators.
     */
    public class Translator : GLib.Object {

        /* The query template. */
        private const string QUERY = """
SELECT rolas.path
FROM rolas
JOIN performers ON rolas.id_performer = performers.id_performer
JOIN albums ON rolas.id_album = albums.id_album
WHERE THE_WHERE
""";

        /* The tokenizer. */
        private Tokenizer tokenizer;
        /* The analyzer. */
        private Analyzer analyzer;

        /**
         * Sets up a translator.
         */
        public Translator() {
            tokenizer = new Tokenizer();
            analyzer = new Analyzer();
        }

        /**
         * Translates the input string to SQL.
         * @param input the input string.
         * @return a SQL translation of the input string.
         */
        public string translate(string input) throws SyntaxError {
            var tokens = tokenizer.tokenize(input);
            var tree = analyzer.analyze(tokens);
            var where = do_translation(tree);
            return QUERY.replace("THE_WHERE", where);
        }

        /* Recursive translation. */
        private string do_translation(Tree tree) {
            if (tree == null)
                return "";
            switch (tree.token.kind) {
            case Token.Kind.COLON:
            case Token.Kind.BINARY_OPERATOR:
                return (" (" +
                        do_translation(tree.left) +
                        tree.token.to_sql() +
                        do_translation(tree.right) + " ) ");
            case Token.Kind.UNARY_OPERATOR:
                return (" ( " +
                        tree.token.to_sql() + " ( " +
                        do_translation(tree.right) + " ) ) ");
            case Token.Kind.FIELD:
            case Token.Kind.STRING:
                return tree.token.to_sql();
            default:
                return "";
            }
        }
    }
}
