/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas.Database {

    /**
     * Class for handling the cache.
     */
    public class Cache : GLib.Object {

        /* The cache directory. */
        private static string CACHE_DIR =
            GLib.Path.build_filename(Environment.get_user_cache_dir(), "rolas");

        /**
         * Creates a new cache interface.
         */
        public Cache() {
            if (!GLib.FileUtils.test(CACHE_DIR,
                                     FileTest.EXISTS | FileTest.IS_DIR))
                DirUtils.create_with_parents(CACHE_DIR, 0755);
        }

        /* Gets the key path. */
        private string get_path(string key) {
            var basename = "%08X".printf(key.hash()) + ".jpg";
            return Path.build_filename(CACHE_DIR, basename);
        }

        /**
         * Whether the cached key exists.
         * @return ''true'' if the key is cached; ''false'' otherwise.
         */
        public bool exists(string key) {
            var path = get_path(key);
            return GLib.FileUtils.test(
                path, FileTest.EXISTS|FileTest.IS_REGULAR);
        }

        /**
         * Return the cached key data.
         * @return the cached key data if the key is cached; an empty byte array
         *         otherwise.
         */
        public uint8[] get_cache_data(string key) {
            uint8[] r = {};
            if (!exists(key))
                return r;
            var path = get_path(key);
            try {
                uint8[] d;
                if (!GLib.FileUtils.get_data(path, out d))
                    return r;
                return d;
            } catch (GLib.Error error) {
                GLib.warning("Could not get data for '$(key)'");
                return r;
            }
        }

        /**
         * Caches the key data.
         * @param key the key to cache.
         * @param data the key data.
         */
        public void set_cache_data(string key, uint8[] data) {
            var path = get_path(key);
            var file = GLib.File.new_for_path(path);
            try {
                GLib.FileIOStream ios =
                    file.create_readwrite(FileCreateFlags.PRIVATE);
                GLib.FileOutputStream fout =
                    ios.output_stream as GLib.FileOutputStream;
                size_t written;
                if (!fout.write_all(data, out written))
                    GLib.warning("Could not set data for '$(key)'");
            } catch (GLib.Error error) {
                GLib.warning("Could not set data for '$(key)'");
            }
        }

        /**
         * Caches the key data on the path.
         * @param key the key to cache.
         * @param path the path with the key data.
         */
        public void set_cache_data_from_path(string key, string path) {
            try {
                uint8[] d;
                if (!GLib.FileUtils.get_data(path, out d))
                    GLib.warning("Could not get data from '$(path)'");
                set_cache_data(key, d);
            } catch (GLib.Error error) {
                GLib.warning("Could not get data from '$(path)'");
            }
        }
    }
}
