/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas.Database {

    /**
     * Enumeration for performer types.
     */
    public enum PerformerType {
        PERSON,
        GROUP,
        UNKNOWN;
    }

    /**
     * Class for database objects.
     */
    public class Database : GLib.Object {

        /* Singleton. */
        private static Database the_database;

        /* The types table. */
        private const string TYPES_TABLE = """
CREATE TABLE types (
    id_type       INTEGER PRIMARY KEY,
    description   TEXT
);
""";

        /* The types data. */
        private const string TYPES_DATA = """
INSERT INTO types VALUES(0, 'Person');
INSERT INTO types VALUES(1, 'Group');
INSERT INTO types VALUES(2, 'Unknown');
""";

        /* The performers table. */
        private const string PERFORMERS_TABLE = """
CREATE TABLE performers (
    id_performer  INTEGER PRIMARY KEY,
    id_type       INTEGER,
    name          TEXT,
    FOREIGN KEY   (id_type) REFERENCES types(id_type)
);
""";
        
        /* The persons table. */
        private const string PERSONS_TABLE = """
CREATE TABLE persons (
    id_person     INTEGER PRIMARY KEY,
    stage_name    TEXT,
    real_name     TEXT,
    birth_date    TEXT,
    death_date    TEXT
);
""";

        /* The groups table. */
        private const string GROUPS_TABLE = """
CREATE TABLE groups (
    id_group      INTEGER PRIMARY KEY,
    name          TEXT,
    start_date    TEXT,
    end_date      TEXT
);
""";

        /* The in group relation table. */
        private const string IN_GROUP_TABLE = """
CREATE TABLE in_group (
    id_person     INTEGER,
    id_group      INTEGER,
    PRIMARY KEY   (id_person, id_group),
    FOREIGN KEY   (id_person) REFERENCES persons(id_person)
    FOREIGN KEY   (id_group) REFERENCES groups(id_group)
);
""";

        /* The albums table. */
        private const string ALBUMS_TABLE = """
CREATE TABLE albums (
    id_album      INTEGER PRIMARY KEY,
    path          TEXT,
    name          TEXT,
    year          INTEGER
);
""";

        /* The albums data. */
        private const string ALBUMS_DATA = """
INSERT INTO albums VALUES (0, '', 'Unknown', 0);
""";

        /* The rolas table. */
        private const string ROLAS_TABLE = """
CREATE TABLE rolas (
    id_rola       INTEGER PRIMARY KEY,
    id_performer  INTEGER,
    id_album      INTEGER,
    path          TEXT,
    title         TEXT,
    track         INTEGER,
    year          INTEGER,
    genre         TEXT,
    FOREIGN KEY   (id_performer) REFERENCES performers(id_performer)
    FOREIGN KEY   (id_album) REFERENCES albums(id_album)
);
""";

        /* Base name of the database. */
        private const string BASENAME = "rolas.db";
        /* Directory name of the database. */
        private static string dirname;
        /* File name of the database. */
        private static string filename;

        /* The database connection. */
        private Sqlite.Database db;
        /* The rolas map. */
        private Gee.TreeMap<string, Rola> rolas;
        /* The translator. */
        private Compiler.Translator translator;

        /* Sets up a new database. */
        private Database() {
            construct_filename();
            verify_filename();
            rolas = new Gee.TreeMap<string, Rola>();
            translator = new Compiler.Translator();

            int rc;
            if ((rc = Sqlite.Database.open(filename, out db)) == 1) {
                GLib.warning("Can't open database: %s", db.errmsg());
                db = null;
            }
        }

        /* Constructs the file name of the database. */
        private void construct_filename() {
            dirname = string.join(Path.DIR_SEPARATOR_S,
                                  Environment.get_home_dir(),
                                  ".local", "share", "rolas");
            filename = string.join(Path.DIR_SEPARATOR_S,
                                   dirname, BASENAME);
        }

        /* Verifies the file name of the database. */
        private void verify_filename() {
            if (!GLib.FileUtils.test(dirname,
                                     FileTest.EXISTS | FileTest.IS_DIR))
                DirUtils.create_with_parents(dirname, 0755);

            if (!GLib.FileUtils.test(filename,
                                     FileTest.EXISTS | FileTest.IS_REGULAR))
                init_database();
        }

        /* Initializes the database. */
        private void init_database() {
            int rc;
            Sqlite.Database db;

            if ((rc = Sqlite.Database.open(filename, out db)) != Sqlite.OK) {
                GLib.warning("Can't open database: %s", db.errmsg());
                db = null;
                return;
            }

            if ((rc = db.exec(TYPES_TABLE)) != Sqlite.OK)
                GLib.warning("Can't create types table: %s", db.errmsg());
            if ((rc = db.exec(TYPES_DATA)) != Sqlite.OK)
                GLib.warning("Can't populate types table: %s", db.errmsg());
            if ((rc = db.exec(PERFORMERS_TABLE)) != Sqlite.OK)
                GLib.warning("Can't create performers table: %s", db.errmsg());
            if ((rc = db.exec(PERSONS_TABLE)) != Sqlite.OK)
                GLib.warning("Can't create persons table: %s", db.errmsg());
            if ((rc = db.exec(GROUPS_TABLE)) != Sqlite.OK)
                GLib.warning("Can't create groups table: %s", db.errmsg());
            if ((rc = db.exec(IN_GROUP_TABLE)) != Sqlite.OK)
                GLib.warning("Can't create in group table: %s", db.errmsg());
            if ((rc = db.exec(ALBUMS_TABLE)) != Sqlite.OK)
                GLib.warning("Can't create albums table: %s", db.errmsg());
            if ((rc = db.exec(ALBUMS_DATA)) != Sqlite.OK)
                GLib.warning("Can't populate albums table: %s", db.errmsg());
            if ((rc = db.exec(ROLAS_TABLE)) != Sqlite.OK)
                GLib.warning("Can't create rolas table: %s", db.errmsg());

            db = null;
        }

        /* Queries the id of a performer. */
        private int query_id_performer(string name) {
            Sqlite.Statement stmt;
            string select;
            int rc = 0;

            select = "SELECT id_performer FROM performers WHERE " +
                @"name = '$(name)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return -1;
            }

            int id_performer = -1;
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    id_performer = stmt.column_int(0);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return id_performer;
        }

        /* Inserts a performer. */
        private int insert_performer(string name) {
            string insert = "INSERT INTO performers(id_type, name) " +
                @"VALUES (2, '$(name)')";
            if (db.exec(insert) != Sqlite.OK) {
                GLib.warning(@"Error inserting performer: '$(name)'");
                return -1;
            }
            return query_id_performer(name);
        }

        /* Gets a performer id. */
        private int get_id_performer(Rola rola) {
            var artist = sql_escape(rola.artist);
            int id_performer = query_id_performer(artist);
            if (id_performer == -1)
                return insert_performer(artist);
            return id_performer;
        }

        /* Queries the id of an album. */
        private int query_id_album(string name, string path) {
            Sqlite.Statement stmt;
            string select;
            int rc = 0;

            select = "SELECT id_album FROM albums WHERE " +
                @"name = '$(name)' AND path = '$(path)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return -1;
            }

            int id_album = -1;
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    id_album = stmt.column_int(0);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return id_album;
        }

        /* Inserts an album. */
        private int insert_album(string name, string path, int year) {
            string insert = "INSERT INTO albums(name, path, year) " +
                @"VALUES ('$(name)', '$(path)', $(year))";
            if (db.exec(insert) != Sqlite.OK) {
                GLib.warning(@"Error inserting album: '$(name)'");
                return -1;
            }
            return query_id_album(name, path);
        }

        /* Gets an album id. */
        private int get_id_album(Rola rola) {
            var file = GLib.File.new_for_path(rola.path);
            var p = file.get_parent();
            var dir = (p == null) ? "" : p.get_path();

            var album = sql_escape(rola.album);
            int id_album = query_id_album(album, dir);
            if (id_album == -1)
                return insert_album(album, dir, rola.year);
            return id_album;
        }

        /* Queries the id of a rola. */
        private int query_id_rola(string path) {
            Sqlite.Statement stmt;
            string select;
            int rc = 0;

            select = @"SELECT id_rola FROM rolas WHERE path = '$(path)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return -1;
            }

            int id_rola = -1;
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    id_rola = stmt.column_int(0);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return id_rola;
        }

        /**
         * Adds a rola to the database.
         * @param rola the rola to add.
         */
        public void add_rola(Rola rola) {
            int id_performer = get_id_performer(rola);
            int id_album = get_id_album(rola);
            if (id_album == -1 || id_performer == -1) {
                GLib.warning(@"Could not add rola: '$(rola.artist)'");
                return;
            }
            string title = sql_escape(rola.title);
            string insert = "INSERT INTO rolas (" +
                "    id_performer, id_album, path, title, " +
                "    track, year, genre) VALUES (" +
                @"$(id_performer), $(id_album), '$(rola.path)', " +
                @"'$(title)', '$(rola.track)', $(rola.year)," +
                @"'$(rola.genre)')";

            if (db.exec(insert) != Sqlite.OK)
                GLib.warning(@"Error inserting rola: '$(rola.title)'");

            rolas[rola.path] = rola;
        }

        /**
         * Loads all the rolas from the database.
         * @return  all the rolas from the database.
         */
        public Gee.TreeMap<string, Rola> load_rolas() {
            rolas.clear();

            Sqlite.Statement stmt;
            string select;
            int rc = 0;

            select = "SELECT rolas.path, rolas.title, performers.name, " +
                "albums.name, rolas.track, rolas.year, rolas.genre " +
                "FROM rolas JOIN performers ON " +
                "rolas.id_performer = performers.id_performer " +
                "JOIN albums ON rolas.id_album = albums.id_album " +
                "ORDER BY rolas.path";
                
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return rolas;
            }
            
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    var path = stmt.column_text(0);
                    var title = sql_unescape(stmt.column_text(1));
                    var artist = sql_unescape(stmt.column_text(2));
                    var album = sql_unescape(stmt.column_text(3));
                    var track = stmt.column_int(4);
                    var year = stmt.column_int(5);
                    var genre = sql_unescape(stmt.column_text(6));
                    var rola = new Rola.from_data(
                        path, title, artist, album, track, year, genre);
                    rolas[path] = rola;
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return rolas;
        }

        /**
         * Updates an album to a new name and year.
         * @param rola the rola that defines the album.
         * @param name the new name of the album.
         * @param year the new year of the album.
         */
        public Gee.Set<Rola> update_album(Rola rola, string name, int year) {
            var modified = new Gee.TreeSet<Rola>();
            int id_album = get_id_album(rola);

            string update =
                @"UPDATE albums SET name = '$(name)', year = $(year) " +
                @"WHERE id_album = $(id_album)";
            if (db.exec(update) != Sqlite.OK) {
                GLib.warning(@"Error updating album '$(name)'");
                return modified;
            }

            Sqlite.Statement stmt;
            int rc = 0;

            string select = "SELECT path FROM rolas WHERE " +
                @"id_album = $(id_album)";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return modified;
            }

            var paths = new Gee.TreeSet<string>();
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    paths.add(stmt.column_text(0));
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            foreach (var path in paths) {
                var r = rolas[path];
                r.album = name;
                r.year = year;
                modified.add(r);
            }

            return modified;
        }

        /**
         * Returns the performer type of a rola.
         * @param rola the rola.
         * @return the performer type of a rola.
         */
        public PerformerType get_performer_type(Rola rola) {
            Sqlite.Statement stmt;
            int rc = 0;
            PerformerType pt = PerformerType.UNKNOWN;

            string select = "SELECT id_type FROM performers WHERE " +
                @"name = '$(rola.artist)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return pt;
            }

            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    pt = (PerformerType)stmt.column_int(0);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return pt;
        }

        /**
         * Retrieves a person from the database.
         * @param stage_name the stage name of the person.
         * @param person where to store the person.
         * @return ''true'' if the person is retrieved; ''false'' otherwise.
         */
        public bool get_person(string stage_name, out Person person) {
            person = null;
            Sqlite.Statement stmt;
            int rc = 0;

            string select = "SELECT * FROM persons WHERE " +
                @"stage_name = '$(stage_name)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return false;
            }

            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    var id_person =  stmt.column_int(0);
                    var real_name = stmt.column_text(2);
                    var birth_date = stmt.column_text(3);
                    var death_date = stmt.column_text(4);
                    person = new Person(id_person, stage_name, real_name,
                                        birth_date, death_date);
                    return true;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return false;
        }

        /**
         * Retrieves a group from the database.
         * @param stage_name the stage name of the group.
         * @param group where to store the group.
         * @return ''true'' if the group is retrieved; ''false'' otherwise.
         */
        public bool get_group(string name, out Group group) {
            group = null;
            Sqlite.Statement stmt;
            int rc = 0;

            string select = @"SELECT * FROM groups WHERE name = '$(name)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return false;
            }

            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    var id_group = stmt.column_int(0);
                    var start_date = stmt.column_text(2);
                    var end_date = stmt.column_text(3);
                    group = new Group(id_group, name, start_date, end_date,
                                      get_group_members(id_group));
                    return true;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return false;
        }

        /**
         * Retrieves the members of a group from the database.
         * @param id_group the group id.
         * @return the members of a group from the database.
         */
        private Gee.TreeSet<Person> get_group_members(int id_group) {
            var ms = new Gee.TreeSet<Person>();
            Sqlite.Statement stmt;
            int rc = 0;

            string select = """SELECT * FROM persons
                               WHERE id_person IN (
                                   SELECT id_person FROM in_group
                                   WHERE id_group = ID_GROUP)""";
            select = select.replace("ID_GROUP", @"$(id_group)");
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return ms;
            }

            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    var id = stmt.column_int(0);
                    var sn = stmt.column_text(1);
                    var rn = stmt.column_text(2);
                    var bd = stmt.column_text(3);
                    var dd = stmt.column_text(4);
                    ms.add(new Person(id, sn, rn, bd, dd));
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return ms;
        }

        /**
         * Updates a person with a new real name, birth date and death date.
         * @param rola the rola tha defines the person.
         * @param real_name the real name of the person.
         * @param birth_date the birth date of the person.
         * @param death_date the death date of the person.
         */
        public void update_person(Rola rola, string real_name,
                                  string birth_date, string death_date) {
            Sqlite.Statement stmt;
            int rc = 0;

            string select = "SELECT id_person FROM persons WHERE " +
                @"stage_name = '$(rola.artist)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return;
            }

            int id_person = -1;
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    id_person = stmt.column_int(0);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            if (id_person == -1)
                insert_persons_table(rola.artist, real_name,
                                     birth_date, death_date);
            else
                update_persons_table(id_person, real_name,
                                     birth_date, death_date);

            update_performer_type(rola, PerformerType.PERSON);
        }

        /* Inserts a person in persons. */
        private void insert_persons_table(string stage_name,
                                          string real_name,
                                          string birth_date,
                                          string death_date) {
            string insert = "INSERT INTO persons (stage_name, real_name, " +
                @"birth_date, death_date) VALUES ('$(stage_name)', " +
                @"'$(real_name)', '$(birth_date)', '$(death_date)')";
            if (db.exec(insert) != Sqlite.OK)
                GLib.warning(@"Error inserting person: '$(stage_name)'");
        }

        /* Updates a person in persons. */
        private void update_persons_table(int id_person,
                                          string real_name,
                                          string birth_date,
                                          string death_date) {
            string update =
                @"UPDATE persons SET real_name = '$(real_name)', " +
                @"birth_date = '$(birth_date)', " +
                @"death_date = '$(death_date)' " +
                @"WHERE id_person = $(id_person)";
            if (db.exec(update) != Sqlite.OK)
                GLib.warning(@"Error updating person '$(real_name)'");
        }

        /**
         * Adds a person to the database.
         * @param stage_name the stage name of the person.
         * @param real_name the real name of the person.
         * @param birth_date the birth date of the person.
         * @param death_date the death date of the person.
         * @return the person added to the database.
         */
        public Person add_person(string stage_name,
                                 string real_name,
                                 string birth_date,
                                 string death_date) {
            insert_persons_table(stage_name, real_name,
                                 birth_date, death_date);
            Person person;
            get_person(stage_name, out person);
            return person;
        }

        /**
         * Updates a group with a new start date and end date.
         * @param rola the rola tha defines the group.
         * @param start_date the start date of the group.
         * @param end_date the end date of the group.
         */
        public void update_group(Rola rola, string start_date,
                                 string end_date) {
            Sqlite.Statement stmt;
            int rc = 0;

            string select = "SELECT id_group FROM groups WHERE " +
                @"name = '$(rola.artist)'";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return;
            }

            int id_group = -1;
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    id_group = stmt.column_int(0);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            if (id_group == -1)
                insert_groups_table(rola.artist, start_date, end_date);
            else
                update_groups_table(id_group, rola.artist,
                                    start_date, end_date);

            update_performer_type(rola, PerformerType.GROUP);
        }

        /* Inserts a group in groups. */
        private void insert_groups_table(string name, string start_date,
                                         string end_date) {
            
            string insert = "INSERT INTO groups (name, " +
                @"start_date, end_date) VALUES ('$(name)', " +
                @"'$(start_date)', '$(end_date)')";
            if (db.exec(insert) != Sqlite.OK)
                GLib.warning(@"Error inserting group: '$(name)'");
        }

        /* Updates a group in groups. */
        private void update_groups_table(int id_group, string name,
                                         string start_date,
                                         string end_date) {
            string update =
                @"UPDATE groups SET start_date = '$(start_date)', " +
                @"end_date = '$(end_date)' WHERE id_group = $(id_group)";
            if (db.exec(update) != Sqlite.OK)
                GLib.warning(@"Error updating group '$(name)'");
        }

        /* Updates a performer type. */
        private void update_performer_type(Rola rola, PerformerType pt) {
            int t = (int)pt;
            string update =
                @"UPDATE performers SET id_type = $(t) " +
                @"WHERE name = '$(rola.artist)'";
            if (db.exec(update) != Sqlite.OK)
                GLib.warning(@"Error updating performer '$(rola.artist)'");
        }

        /**
         * Returns the group id of a rola.
         * @param rola the rola that defines the group.
         * @return the group id of a rola, or -1 if there is an error.
         */
        public int get_group_id(Rola rola) {
            int id_rola = query_id_rola(rola.path);
            Sqlite.Statement stmt;
            int rc = 0;

            string select = "SELECT groups.id_group FROM "+
              "rolas JOIN performers ON " +
              "rolas.id_performer = performers.id_performer " +
              "JOIN groups ON performers.name = groups.name " +
              @"WHERE rolas.id_rola = $(id_rola)";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return -1;
            }

            int id_group = -1;
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    id_group = stmt.column_int(0);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return id_group;
        }

        /**
         * Loads the persons from the database.
         * @return a set of persons from the database.
         */
        public Gee.TreeSet<Person> load_persons() {
            Sqlite.Statement stmt;
            string select;
            int rc = 0;

            var persons = new Gee.TreeSet<Person>();
            select = "SELECT * FROM persons";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return persons;
            }
            
            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    var id_person = stmt.column_int(0);
                    var stage_name = sql_unescape(stmt.column_text(1));
                    var real_name = sql_unescape(stmt.column_text(2));
                    var birth_date = sql_unescape(stmt.column_text(3));
                    var death_date = sql_unescape(stmt.column_text(4));
                    var person = new Person(id_person, stage_name,
                                            real_name, birth_date,
                                            death_date);
                    persons.add(person);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return persons;
        }

        /**
         * Queries rolas from the database.
         * @param query the query.
         * @return a set of rolas from the database that match the query.
         */
        public Gee.Set<Rola> query_rolas(string query) {
            var qs = new Gee.TreeSet<Rola>();
            Sqlite.Statement stmt;
            string select = null;
            int rc = 0;

            try {
                select = translator.translate(query);
            } catch (Compiler.SyntaxError error) {
                GLib.warning("Translator error: %s", error.message);
                return qs;
            }

            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return qs;
            }

            do {
                rc = stmt.step();
                switch (rc) {
                case Sqlite.DONE:
                    break;
                case Sqlite.ROW:
                    var path = stmt.column_text(0);
                    qs.add(rolas[path]);
                    break;
                default:
                    GLib.warning("Error: %d, %s", rc, db.errmsg());
                    break;
                }
            } while (rc == Sqlite.ROW);

            return qs;
        }

        /* Whether a person is in a group. */
        private bool person_in_group(Person person, int id_group) {
            Sqlite.Statement stmt;
            int rc = 0;

            string select = "SELECT * FROM in_group WHERE " +
                @"id_person = $(person.id) AND id_group = $(id_group)";
            if ((rc = db.prepare_v2(select, -1, out stmt, null)) == 1) {
                GLib.warning("SQL error: %d, %s", rc, db.errmsg());
                return false;
            }

            return stmt.step() == Sqlite.ROW;
        }

        /**
         * Adds a person to a group.
         * @param person the person to add.
         * @param rola the rola that defines the group.
         */
        public void add_person_to_group(Person person, Rola rola) {
            int id_group = get_group_id(rola);
            if (person_in_group(person, id_group))
                return;
            string insert =
                @"INSERT INTO in_group VALUES ($(person.id), $(id_group))";
            if (db.exec(insert) != Sqlite.OK)
                GLib.warning("Error inserting person in group: '%s'",
                             person.stage_name);
        }

        /* Escapes a SQL string. */
        private string sql_escape(string s) {
            return s.replace(" & ", " &amp; ").replace("'", "&#039;");
        }

        /* Unescapes a SQL string. */
        private string sql_unescape(string s) {
            return s.replace(" &amp; ", " & ").replace("&#039;", "'");
        }

        /**
         * Returns a singleton with the database.
         */
        public static Database get_database() {
            if (the_database == null)
                the_database = new Database();
            return the_database;
        }
    }
}
