/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas.Database {

    /**
     * Class for a miner.
     */
    public class Miner : GLib.Object {

        /**
         * Signals that the pre-mining is done.
         */
        public signal void pre_mining_done();

        /**
         * Signals that the mining is done.
         */
        public signal void mining_done();

        /* Music directory. Hard coded. */
        private const string MUSIC = "Music";

        /* Already mined rolas.*/
        private Gee.Set<string> skip;
        /* The directories to mine. */
        private Gee.Queue<GLib.File> dirs;
        /* The files to mine. */
        private Gee.TreeSet<GLib.File> files;
        /* The files to mine iterator. */
        private Gee.Iterator<GLib.File> iterator;
        /* Whether to cancel the minin. */
        private bool cancel_mining;
        /* Number of mined rolas.. */
        private int mined;

        /**
         * Sets up a miner.
         * @param skip rola paths to skip.
         */
        public Miner(Gee.Set<string> skip) {
            this.skip = skip;
            var root = string.join(GLib.Path.DIR_SEPARATOR_S,
                                   GLib.Environment.get_home_dir(),
                                   MUSIC);
            dirs = new Gee.LinkedList<GLib.File>();
            files = new Gee.TreeSet<GLib.File>(compare_files);
            var r = GLib.File.new_for_path(root);
            var ft = r.query_file_type(GLib.FileQueryInfoFlags.NONE);
            if (ft != GLib.FileType.DIRECTORY)
                return;
            dirs.add(r);
        }

        /* Premines a directory. */
        private void pre_mine_dir(GLib.File dir) {
            try {
                var enumerator = dir.enumerate_children(
                    "standard::*", FileQueryInfoFlags.NONE);
                GLib.FileInfo info = null;
                while ((info = enumerator.next_file()) != null) {
                    var name = info.get_name();
                    var child = dir.resolve_relative_path(name);
                    switch (info.get_file_type()) {
                    case GLib.FileType.DIRECTORY:
                        dirs.add(child);
                        break;
                    case GLib.FileType.REGULAR:
                        if (name.has_suffix(".mp3") &&
                            !skip.contains(child.get_path()))
                            files.add(child);
                        break;
                    }
                }
            } catch (GLib.Error error) {
                GLib.warning("Error minining %s", dir.get_basename());
            }
        }

        /**
         * Pre-mines the music directory.
         */
        public void pre_mine() {
            while (!dirs.is_empty)
                pre_mine_dir(dirs.poll());
            iterator = files.iterator();
            cancel_mining = false;
            mined = 0;
            pre_mining_done();
        }

        /**
         * Mines one rola.
         * @param rola where to put the rola.
         * @return ''true'' if the rola was mined; ''false'' otherwise.
         */
        public bool mine(out Rola rola) {
            rola = null;
            if (cancel_mining || !iterator.has_next()) {
                mined = 0;
                return false;
            }
            iterator.next();
            var file = iterator.get();
            var path = file.get_path();
            rola = new Rola(path);
            mined++;
            return true;
        }

        /**
         * Cancels the mining.
         */
        public void cancel() {
            cancel_mining = true;
        }

        /**
         * Returns the progress of the mining.
         * @return the progress of the mining.
         */
        public double get_progress() {
            return (mined * 1.0) / files.size;
        }

        /* Compares two files. */
        private static int compare_files(GLib.File f, GLib.File g) {
            return f.get_path().collate(g.get_path());
        }
    }
}
