/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas {

    /**
     * Class for adding members to band dialogs.
     */
    [GtkTemplate (ui = "/mx/unam/ciencias/Rolas/add-member-dialog.ui")]
    public class AddMemberDialog : Gtk.Dialog {

        /* Notebook pages. */
        private enum Page {
            NEW,
            EXISTING;
        }

        /* The persons model. */
        [GtkChild]
        private Gtk.ListStore persons_model;

        /* The add member button. */
        [GtkChild]
        private Gtk.Button add_member_button;

        /* The notebook. */
        [GtkChild]
        private Gtk.Notebook notebook;

        /* The stage name entry. */
        [GtkChild]
        private Gtk.Entry stage_name_entry;

        /* The real name entry. */
        [GtkChild]
        private Gtk.Entry real_name_entry;

        /* The birth date entry. */
        [GtkChild]
        private Gtk.Entry birth_date_entry;

        /* The death date entry. */
        [GtkChild]
        private Gtk.Entry death_date_entry;

        /* The persons combo box. */
        [GtkChild]
        private Gtk.ComboBoxText persons_combo;

        /**
         * The rola.
         */
        public Database.Rola rola {
            get { return _rola; }
            set { define_rola(value); }
        }
        private Database.Rola _rola;

        /* The database. */
        private Database.Database db;

        /**
         * Sets up a new add member dialog.
         * @param window transient window.
         */
        public AddMemberDialog(Gtk.Window window) {
            GLib.Object(use_header_bar: 1);
            transient_for = window;
            db = Database.Database.get_database();
            add_member_button.sensitive = false;
        }

        /**
         * The on close activated callback.
         */
        [GtkCallback]
        public bool on_close_activated() {
            on_cancel_clicked();
            return false;
        }

        /**
         * The on add member clicked callback.
         */
        [GtkCallback]
        public void on_add_member_clicked() {
            hide();
            switch (notebook.page) {
            case Page.NEW:
                create_person();
                break;
            case Page.EXISTING:
                add_member(get_selected_person());
                break;
            }
            response(Gtk.ResponseType.OK);
        }

        /**
         * The on cancel clicked callback.
         */
        [GtkCallback]
        public void on_cancel_clicked() {
            hide();
            response(Gtk.ResponseType.CANCEL);
        }

        /**
         * The on page switched callback.
         */
        [GtkCallback]
        public void on_page_switched(Gtk.Widget page, uint page_num) {
            switch (page_num) {
            case Page.NEW:
                check_person_info();
                break;
            case Page.EXISTING:
                check_persons_combo();
                break;
            }
        }

        /**
         * The on person info changed callback.
         */
        [GtkCallback]
        public void on_person_info_changed() {
            check_person_info();
        }

        /**
         * The on persons combo changed callback.
         */
        [GtkCallback]
        public void on_persons_combo_changed() {
            check_persons_combo();
        }

        /* Checks a person info. */
        private void check_person_info() {
            if (stage_name_entry.text == "" || real_name_entry.text== "") {
                add_member_button.sensitive = false;
                return;
            }
            add_member_button.sensitive = true;
        }

        /* Checks the persons combo. */
        private void check_persons_combo() {
            Gtk.TreeIter iter;
            if (!persons_combo.get_active_iter(out iter)) {
                add_member_button.sensitive = false;
                return;
            }
            add_member_button.sensitive = true;
        }

        /* Returns the selected person. */
        private Database.Person get_selected_person() {
            Database.Person person = null;
            Gtk.TreeIter iter;
            if (persons_combo.get_active_iter(out iter))
                persons_model.get(iter, 0, out person);
            return person;
        }

        /* Creates a person. */
        private void create_person() {
            var person = db.add_person(stage_name_entry.text,
                                       real_name_entry.text,
                                       birth_date_entry.text,
                                       death_date_entry.text);
            add_member(person);
        }

        /* Adds a member. */
        private void add_member(Database.Person person) {
            db.add_person_to_group(person, rola);
        }

        /* Defines the rola. */
        private void define_rola(Database.Rola rola) {
            _rola = rola;
            var persons = db.load_persons();
            foreach (var person in persons) {
                Gtk.TreeIter iter;
                persons_model.append(out iter);
                persons_model.set(iter,
                                  0, person,
                                  1, person.stage_name);
            }
        }
    }
}
