/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas {

    /**
     * Class for rolas application windows.
     */
    [GtkTemplate (ui = "/mx/unam/ciencias/Rolas/application-window.ui")]
    public class ApplicationWindow : Gtk.ApplicationWindow {

        /* The actions. */
        private enum Action {
            PLAY_STOP,
            MINE,
            PERFORMER,
            ALBUM,
            SEARCH,
            LIST_VIEW,
            GRID_VIEW,
            NONE;
        }

        /* Notebook pages. */
        private enum Page {
            LIST,
            GRID;
        }

        /* The rola columns. */
        private enum Column {
            ROLA,
            TITLE,
            ARTIST,
            ALBUM,
            GENRE;
        }

        /* Default subtitle. */
        private const string SUBTITLE = "Music Database";
        /* Application menu. */
        private const string APP_MENU =
            "/mx/unam/ciencias/Rolas/application-menu.ui";
        /* Icon. */
        private const string ICON =
            "/mx/unam/ciencias/Rolas/rolas.svg";

        /* The header bar. */
        [GtkChild]
        private Gtk.HeaderBar header;

        /* The mine button. */
        [GtkChild]
        private Gtk.Button mine_button;

        /* The play button. */
        [GtkChild]
        private Gtk.Button play_button;

        /* The play image. */
        [GtkChild]
        private Gtk.Image play_image;

        /* The list view toggle button. */
        [GtkChild]
        private Gtk.ToggleButton list_view_toggle_button;

        /* The grid view menu button. */
        [GtkChild]
        private Gtk.ToggleButton grid_view_toggle_button;

        /* The filter entry. */
        [GtkChild]
        private Gtk.Entry filter_entry;

        /* The menu button. */
        [GtkChild]
        private Gtk.MenuButton app_menu_button;

        /* The notebook. */
        [GtkChild]
        private Gtk.Notebook notebook;

        /* The tree view. */
        [GtkChild]
        private Gtk.TreeView rolas_tree_view;

        /* The tree selection. */
        [GtkChild]
        private Gtk.TreeSelection rolas_tree_selection;

        /* The rolas list scrolled window. */
        [GtkChild]
        Gtk.ScrolledWindow rolas_list_scroll;

        /* The rolas grid scrolled window. */
        [GtkChild]
        Gtk.ScrolledWindow rolas_grid_scroll;

        /* The tree view. */
        [GtkChild]
        private Gtk.FlowBox rolas_flow_box;

        /* The filtered rolas model. */
        [GtkChild]
        private Gtk.TreeModelFilter filtered_rolas_model;

        /* The rolas model. */
        [GtkChild]
        private Gtk.ListStore rolas_model;

        /* The rola image. */
        [GtkChild]
        private Gtk.Image rola_image;

        /* The rola image. */
        [GtkChild]
        private Gtk.Label play_time;

        /* The rola title. */
        [GtkChild]
        private Gtk.Label rola_title;

        /* The rola album. */
        [GtkChild]
        private Gtk.Label rola_album;

        /* The rola artist. */
        [GtkChild]
        private Gtk.Label rola_artist;

        /* The play adjustment artist. */
        [GtkChild]
        private Gtk.Adjustment play_adjustment;

        /* The Rolas application. */
        private Application rolas_app;
        /* The roals store. */
        private GLib.ListStore rolas_store;
        /* The shortcuts window. */
        private ShortcutsWindow shortcuts;
        /* The edit performer dialog. */
        private EditPerformerDialog edit_performer_dialog;
        /* The edit album dialog. */
        private EditAlbumDialog edit_album_dialog;
        /* The media player*/
        private Player player;
        /* The rolas.*/
        private Gee.TreeMap<string, Database.Rola> rolas;
        /* The paths iterator. */
        private Gee.Iterator<string> paths_iterator;
        /* The query set. */
        private Gee.Set<Database.Rola> query_set;
        /* The paths counter. */
        private int paths_counter;
        /* The database. */
        private Database.Database db;
        /* The miner. */
        private Database.Miner miner;
        /* The selected rola. */
        private Database.Rola rola;
        /* Cancel loading flag. */
        private bool cancel_loading;
        /* Cancel mining flag. */
        private bool cancel_mining;
        /* Cancel playing flag. */
        private bool playing;

        /**
         * Defines the initial state of an application window.
         * @param application the Rolas application.
         */
        public ApplicationWindow(Application application) {
            GLib.Object(application: application);
            this.rolas_app = application;

            var builder = new Gtk.Builder.from_resource(APP_MENU);
            var app_menu = builder.get_object("menu") as GLib.MenuModel;
            app_menu_button.set_menu_model(app_menu);

            filtered_rolas_model.set_visible_func(rola_is_visible);

            list_view_toggle_button.bind_property(
                "active",
                grid_view_toggle_button,
                "active",
                GLib.BindingFlags.INVERT_BOOLEAN|
                GLib.BindingFlags.BIDIRECTIONAL);

            rolas_store = new GLib.ListStore(typeof(Database.Rola));
            rolas_flow_box.bind_model(rolas_store, create_rola_view);

            rola_image.pixbuf = RolaView.default_cover();

            rolas_flow_box.set_vadjustment(rolas_grid_scroll.vadjustment);
            rolas_flow_box.set_hadjustment(rolas_grid_scroll.hadjustment);

            db = Database.Database.get_database();
            rolas = db.load_rolas();
            if (rolas.size > 0) {
                mine_button.sensitive = false;
                header.subtitle = "Loading (0.0%%)";
                paths_iterator = rolas.keys.iterator();
                paths_counter = 0;
                GLib.Idle.add(load_rolas);
            }
        }

        /**
         * The on shortcuts activated callback.
         */
        public void on_shortcuts_activated() {
            if (shortcuts == null)
                shortcuts = new ShortcutsWindow();
            shortcuts.show_all();
        }

        /**
         * The on about activated callback.
         */
        public void on_about_activated() {
            Gdk.Pixbuf icon = null;
            try {
                icon = new Gdk.Pixbuf.from_resource_at_scale(
                    ICON, 256, 256, true);
            } catch (GLib.Error error) {
                GLib.warning(@"Error loading icon: $(error.message))");
            }
            Gtk.show_about_dialog(
                this,
                "authors", new string[] { "Canek Peláez Valdés " +
                                          "<canek@ciencias.unam.mx>" },
                "comments", "Music Database",
                "copyright", "Copyright © 2018 Canek Peláez Valdés",
                "license-type", Gtk.License.GPL_3_0,
                "logo", icon,
                "version", "0.1",
                "website",
                "https://aztlan.fciencias.unam.mx/gitlab/canek/rolas",
                "website-label", "Project homepage",
                "wrap-license", true);
        }

        /**
         * The on quit activated callback.
         */
        [GtkCallback]
        public void on_quit_activated() {
            if (player != null && playing)
                set_status_idle();
            cancel_loading = true;
            cancel_mining = true;
        }

        /* The on key press event callback. */
        [GtkCallback]
        private bool on_key_press_event(Gdk.EventKey event) {
            switch (get_action(event)) {
            case Action.PLAY_STOP:
                on_play_clicked();
                return true;
            case Action.MINE:
                on_mine_clicked();
                return true;
            case Action.PERFORMER:
                on_performer_clicked();
                return true;
            case Action.ALBUM:
                on_album_clicked();
                return true;
            case Action.SEARCH:
                filter_entry.grab_focus();
                return true;
            case Action.LIST_VIEW:
                list_view_toggle_button.active = true;
                return true;
            case Action.GRID_VIEW:
                grid_view_toggle_button.active = true;
                return true;
            default:
                return false;
            }
        }

        /**
         * The on play clicked callback.
         */
        [GtkCallback]
        public void on_play_clicked() {
            if (playing)
                set_status_idle();
            else
                set_status_playing();
        }

        /**
         * The on mine clicked callback.
         */
        [GtkCallback]
        public void on_mine_clicked() {
            miner = new Database.Miner(rolas.keys);
            miner.pre_mining_done.connect(() => GLib.Idle.add(do_mine));
            new GLib.Thread<void*>("miner", do_pre_mine);
            mine_button.sensitive = false;
        }

        /**
         * The on filter entry activated callback.
         */
        [GtkCallback]
        public void on_filter_entry_activated() {
            var query = filter_entry.text;
            if (query == "") {
                query_set = null;
            } else {
                query_set = db.query_rolas(query);
            }
            filtered_rolas_model.refilter();
        }

        /**
         * The on main view toggled callback.
         */
        [GtkCallback]
        public void on_main_view_toggled() {
            if (list_view_toggle_button.active)
                notebook.page = Page.LIST;
            else
                notebook.page = Page.GRID;
        }

        /**
         * The on performer activated callback.
         */
        [GtkCallback]
        public void on_performer_clicked() {
            var r = get_selected_rola();
            if (r == null)
                return;
            if (edit_performer_dialog == null)
                edit_performer_dialog = new EditPerformerDialog(this);
            edit_performer_dialog.rola = r;
            edit_performer_dialog.run();
        }

        /**
         * The on album activated callback.
         */
        [GtkCallback]
        public void on_album_clicked() {
            var r = get_selected_rola();
            if (r == null)
                return;
            if (edit_album_dialog == null)
                edit_album_dialog = new EditAlbumDialog(this);
            edit_album_dialog.rola = r;
            int rc = edit_album_dialog.run();
            if (rc != Gtk.ResponseType.OK)
                return;
            var changed = edit_album_dialog.get_changed_rolas();
            rolas_model.@foreach((model, path, iter) => {
                    model.get(iter, 0, out r);
                    if (r in changed) {
                        var a = Database.Util.sanitize(r.album);
                        rolas_model.set(iter, Column.ALBUM, a);
                    }
                    return false;
                });
        }

        /**
         * The on rola selection changed callback.
         */
        [GtkCallback]
        private void on_rola_selection_changed() {
            var r = get_selected_rola();
            if (r == null) {
                if (!playing)
                    play_button.sensitive = false;
            } else {
                select_other_view(r);
                if (!playing) {
                    rola = r;
                    update_rola_info();
                    play_button.sensitive = true;
                }
            }
        }

        /**
         * The on rola list activated callback.
         * @param path the activated path.
         * @param column the activated column.
         */
        [GtkCallback]
        public void on_rola_list_activated(Gtk.TreePath path,
                                           Gtk.TreeViewColumn column) {
            on_rola_activated();
        }

        /**
         * The on rola grid activated callback.
         * @param child the activated child.
         */
        [GtkCallback]
        public void on_rola_grid_activated(Gtk.FlowBoxChild child) {
            on_rola_activated();
        }

        /**
         * The on scale change value callback.
         * @param scroll the scroll type.
         * @param value the value.
         * @return ''true'' if the player cannot seek the value, ''false''
         *         otherwise.
         */
        [GtkCallback]
        public bool on_scale_change_value(Gtk.ScrollType scroll, double value) {
            if (!playing)
                return true;
            if (value < 0.0)
                value = 0.0;
            if (value > 1.0)
                value = 1.0;
            if (!player.seek(value)) {
                set_status_idle();
                return true;
            }
            return false;
        }

        /* Updates the current rola info. */
        private void update_rola_info() {
            rola_title.label = rola.title;
            rola_album.label = rola.album;
            rola_artist.label = rola.artist;
            rola_image.pixbuf = RolaView.get_cover(rola);
        }

        /* On rola activated mixed callback. */
        private void on_rola_activated() {
            if (playing) {
                set_status_idle();
                rola = get_selected_rola();
                update_rola_info();
                GLib.Idle.add(change_played_song);
            } else {
                set_status_playing();
            }
        }

        /* Load rolas idle function. */
        private bool load_rolas() {
            if (cancel_loading)
                return false;
            if (!paths_iterator.has_next()) {
                mine_button.sensitive = true;
                header.subtitle = SUBTITLE;
                return false;
            }
            paths_iterator.next();
            var path = paths_iterator.get();
            var rola = rolas[path];
            double p = (++paths_counter * 100.0) / rolas.size;
            header.subtitle = "Loading (%2.2f%%)".printf(p);
            add_rola(rola);
            return true;
        }

        /* Monitor player idle function. */
        private bool monitor_player() {
            if (!playing)
                return false;
            int64 d, p;
            double f = player.get_completion(out d, out p);
            play_adjustment.set_value(f);
            int64 t = p / 1000000000l;
            int m = (int)(t / 60);
            int s = (int)(t % 60);
            play_time.label = "%02d:%02d".printf(m, s);
            if (d <= p)
                set_status_idle();
            return true;
        }

        /* Change played song idle function. */
        private bool change_played_song() {
            set_status_playing();
            return false;
        }

        /* Starts the player. */
        private void set_status_playing() {
            playing = true;
            play_image.set_from_icon_name("media-playback-stop-symbolic",
                                          Gtk.IconSize.MENU);
            play_button.tooltip_text = "Stop rola";

            player = new Player(rola.path);
            player.play();
            GLib.Idle.add(monitor_player);
        }

        /* Stops the player. */
        private void set_status_idle() {
            player.stop();
            playing = false;
            play_time.label = "00:00";
            play_adjustment.set_value(0.0);
            play_image.set_from_icon_name("media-playback-start-symbolic",
                                          Gtk.IconSize.MENU);
            play_button.tooltip_text = "Play rola";
            var r = get_selected_rola();
            if (r == null) {
                play_button.sensitive = false;
            } else {
                rola = r;
                update_rola_info();
            }
        }

        /* Pre mine thread. */
        private void* do_pre_mine() {
            miner.pre_mine();
            return null;
        }

        /* Adds a rola to every model and collection. */
        private void add_rola(Database.Rola rola) {
            Gtk.TreeIter iter = {};
            rolas_model.append(out iter);
            var title = Database.Util.sanitize(rola.title);
            var artist = Database.Util.sanitize(rola.artist);
            var album = Database.Util.sanitize(rola.album);
            var genre = Database.Util.sanitize(rola.genre);
            rolas_model.set(iter,
                            Column.ROLA,   rola,
                            Column.TITLE,  title,
                            Column.ARTIST, artist,
                            Column.ALBUM,  album,
                            Column.GENRE,  genre);
            rolas_store.append(rola);
        }

        /* Mine idle function. */
        private bool do_mine() {
            if (cancel_mining || miner == null)
                return false;
            Database.Rola rola = null;
            bool keep_mining = miner.mine(out rola);
            double p = miner.get_progress();
            header.subtitle = "Mining (%2.2f%%)".printf(p*100.0);
            if (keep_mining) {
                add_rola(rola);
                db.add_rola(rola);
            } else {
                miner = null;
                mine_button.sensitive = true;
                header.subtitle = SUBTITLE;
            }
            return keep_mining;
        }

        /* Creates a rola view. */
        private Gtk.Widget create_rola_view(GLib.Object object) {
            var rola = (Database.Rola)object;
            return new RolaView(rola);
        }

        /* Selects a rola in the other view. */
        private void select_other_view(Database.Rola rola) {
            switch (notebook.page) {
            case Page.LIST:
                select_flow_box(rola);
                break;
            case Page.GRID:
                select_tree_view(rola);
                break;
            }
        }

        /* Selects a rola in the tree view. */
        private void select_tree_view(Database.Rola rola) {
            filtered_rolas_model.@foreach((model, path, iter) => {
                    Database.Rola r;
                    model.get(iter, 0, out r);
                    if (r == rola) {
                        rolas_tree_selection.select_iter(iter);
                        rolas_tree_view.scroll_to_cell(
                            path, null, false, 0.5f, 0.0f);
                        return true;
                    }
                    return false;
                });
        }

        /* Selects a rola in the flow box. */
        private void select_flow_box(Database.Rola rola) {
            int c = 0;
            var r = (Database.Rola)rolas_store.get_item(c);
            while (r != null) {
                if (r == rola) {
                    Gtk.FlowBoxChild child;
                    child = rolas_flow_box.get_child_at_index(c);
                    rolas_flow_box.select_child(child);
                    var adj = rolas_list_scroll.vadjustment;
                    double a = adj.lower, b = adj.upper, v = adj.value;
                    double gv = (v-a) / (b-a);
                    adj = rolas_grid_scroll.vadjustment;
                    a = adj.lower;
                    b = adj.upper;
                    adj.value = a+(b-a)*gv;
                    return;
                }
                c++;
                var p = rolas_store.get_item(c);
                r = (p != null) ? (Database.Rola)p : null;
            }
        }

        /* Gets the currentle selected rola. */
        private Database.Rola? get_selected_rola() {
            switch (notebook.page) {
            case Page.LIST: return tree_view_selected_rola();
            case Page.GRID: return flow_box_selected_rola();
            }
            return null;
        }

        /* Gets the currentle selected rola in the tree view. */
        private Database.Rola? tree_view_selected_rola() {
            var selection = rolas_tree_selection;
            Gtk.TreeIter iter;
            unowned Gtk.TreeModel model;
            if (selection.get_selected(out model, out iter)) {
                Database.Rola r;
                model.get(iter, 0, out r);
                return r;
            }
            return null;
        }

        /* Gets the currentle selected rola in the flow box. */
        private Database.Rola? flow_box_selected_rola() {
            var list = rolas_flow_box.get_selected_children();
            if (list != null) {
                int index = list.first().data.get_index();
                return (Database.Rola)rolas_store.get_object(index);
            }
            return null;
        }

        /* Whether the iterator is visible or not. */
        private bool rola_is_visible(Gtk.TreeModel model,
                                     Gtk.TreeIter iter) {
            if (query_set == null)
                return true;
            Database.Rola rola;
            model.get(iter, 0, out rola);
            return rola in query_set;
        }

        /* Gets a shortcut. */
        private Action get_action(Gdk.EventKey event) {
            uint key;
            Gdk.keyval_convert_case(event.keyval, null, out key);
            var mod1 = (event.state & Gdk.ModifierType.MOD1_MASK) != 0;
            var control = (event.state & Gdk.ModifierType.CONTROL_MASK) != 0;
            var shift = (event.state & Gdk.ModifierType.SHIFT_MASK) != 0;

            if (key == Gdk.Key.space && !mod1 && control && !shift)
                return Action.PLAY_STOP;
            if (key == Gdk.Key.M && !mod1 && control && !shift)
                return Action.MINE;
            if (key == Gdk.Key.P && !mod1 && control && !shift)
                return Action.PERFORMER;
            if (key == Gdk.Key.L && !mod1 && control && !shift)
                return Action.ALBUM;
            if (key == Gdk.Key.F && !mod1 && control && !shift)
                return Action.SEARCH;
            if (key == Gdk.Key.@1 && !mod1 && control && !shift)
                return Action.LIST_VIEW;
            if (key == Gdk.Key.@2 && !mod1 && control && !shift)
                return Action.GRID_VIEW;
            return Action.NONE;
        }
    }
}
