/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas {

    /**
     * Class for editing performer dialogs.
     */
    [GtkTemplate (ui = "/mx/unam/ciencias/Rolas/edit-performer-dialog.ui")]
    public class EditPerformerDialog : Gtk.Dialog {

        /* The notebook pages. */
        private enum Page {
            PERSON,
            GROUP;
        }

        /* The save button. */
        [GtkChild]
        private Gtk.Button save_button;

        /* The cancel button. */
        [GtkChild]
        private Gtk.Button cancel_button;

        /* The notebook. */
        [GtkChild]
        private Gtk.Notebook notebook;

        /* The stage name entry. */
        [GtkChild]
        private Gtk.Entry stage_name_entry;

        /* The real name entry. */
        [GtkChild]
        private Gtk.Entry real_name_entry;

        /* The birth date entry. */
        [GtkChild]
        private Gtk.Entry birth_date_entry;

        /* The death date entry. */
        [GtkChild]
        private Gtk.Entry death_date_entry;

        /* The name entry. */
        [GtkChild]
        private Gtk.Entry name_entry;

        /* The start date entry. */
        [GtkChild]
        private Gtk.Entry start_date_entry;

        /* The end date entry. */
        [GtkChild]
        private Gtk.Entry end_date_entry;

        /* The show members button. */
        [GtkChild]
        private Gtk.Button show_members_button;

        /* The add member button. */
        [GtkChild]
        private Gtk.Button add_member_button;

        /**
         * The rola.
         */
        public Database.Rola rola {
            get { return _rola; }
            set { define_rola(value); }
        }
        private Database.Rola _rola;

        /* The database. */
        private Database.Database db;
        /* The show members dialog. */
        private ShowMembersDialog show_members_dialog;
        /* The performer type. */
        private Database.PerformerType performer_type;

        /**
         * Sets up a edit performer dialog.
         * @param window the transient window.
         */
        public EditPerformerDialog(ApplicationWindow window) {
            GLib.Object(use_header_bar: 1);
            transient_for = window;
            db = Database.Database.get_database();
            show_members_dialog = new ShowMembersDialog(window);
        }

        /**
         * The on close activated callback.
         * @return ''false'', always. The dialog never gets destroyed.
         */
        [GtkCallback]
        public bool on_close_activated() {
            on_cancel_clicked();
            return false;
        }

        /**
         * The on page switched callback.
         * @param page the page.
         * @param page_num the page number.
         */
        [GtkCallback]
        public void on_page_switched(Gtk.Widget page, uint page_num) {
            switch (page_num) {
            case Page.PERSON:
                save_button.label = "Save person";
                cancel_button.label = "Don't save person";
                performer_type = Database.PerformerType.PERSON;
                break;
            case Page.GROUP:
                save_button.label = "Save group";
                cancel_button.label = "Don't save group";
                performer_type = Database.PerformerType.GROUP;
                break;
            }
        }

        /**
         * The on save clicked callback.
         */
        [GtkCallback]
        public void on_save_clicked() {
            hide();
            switch (performer_type) {
            case Database.PerformerType.PERSON:
                db.update_person(rola,
                                 real_name_entry.text,
                                 birth_date_entry.text,
                                 death_date_entry.text);
                break;
            case Database.PerformerType.GROUP:
                db.update_group(rola,
                                start_date_entry.text,
                                end_date_entry.text);
                break;
            }
            response(Gtk.ResponseType.OK);
        }

        /**
         * The on cancel clicked callback.
         */
        [GtkCallback]
        public void on_cancel_clicked() {
            hide();
            response(Gtk.ResponseType.CANCEL);
        }

        /**
         * The on show members clicked callback.
         */
        [GtkCallback]
        public void on_show_members_clicked() {
            show_members_dialog.run();
        }

        /**
         * The on add members clicked callback.
         */
        [GtkCallback]
        public void on_add_member_clicked() {
            var dialog = new AddMemberDialog(this);
            dialog.rola = _rola;
            dialog.run();
            define_rola_for_group(_rola);
        }

        /* Defines a rola for a person .*/
        private void define_rola_for_person(Database.Rola rola) {
            notebook.get_nth_page(Page.PERSON).visible = true;
            notebook.get_nth_page(Page.GROUP).visible = false;

            stage_name_entry.text = rola.artist;

            Database.Person person;
            if (!db.get_person(rola.artist, out person))
                return;

            real_name_entry.text = person.real_name;
            birth_date_entry.text = person.birth_date;
            death_date_entry.text = person.death_date;
            show_members_button.sensitive = false;
        }

        /* Defines a rola for a group .*/
        private void define_rola_for_group(Database.Rola rola) {
            notebook.get_nth_page(Page.PERSON).visible = false;
            notebook.get_nth_page(Page.GROUP).visible = true;
            add_member_button.sensitive = true;

            name_entry.text = rola.artist;
            Database.Group group;
            if (!db.get_group(rola.artist, out group))
                return;

            start_date_entry.text = group.start_date;
            end_date_entry.text = group.end_date;
            show_members_dialog.group = group;
            show_members_button.sensitive = group.size > 0;
        }

        /* Defines a rola for a person and a group .*/
        private void define_rola_for_both(Database.Rola rola) {
            performer_type = Database.PerformerType.PERSON;
            stage_name_entry.text = rola.artist;
            name_entry.text = rola.artist;
            notebook.get_nth_page(Page.PERSON).visible = true;
            notebook.get_nth_page(Page.GROUP).visible = true;
            show_members_button.sensitive = false;
            add_member_button.sensitive = false;
        }

        /* Defines a rola. */
        private void define_rola(Database.Rola rola) {
            _rola = rola;
            real_name_entry.text = "";
            birth_date_entry.text = "";
            death_date_entry.text = "";
            start_date_entry.text = "";
            end_date_entry.text = "";
            performer_type = db.get_performer_type(rola);
            switch (performer_type) {
            case Database.PerformerType.PERSON:
                define_rola_for_person(rola);
                break;
            case Database.PerformerType.GROUP:
                define_rola_for_group(rola);
                break;
            case Database.PerformerType.UNKNOWN:
                define_rola_for_both(rola);
                break;
            }
        }
    }
}
