/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas {

    /**
     * Class for media players.
     */
    public class Player : GLib.Object {

        /* The pipeline. */
        private Gst.Pipeline pipe;

        /**
         * Creates a player with the given media path.
         * @param path path to the media.
         */
        public Player(string path) {
            try {
                pipe = (Gst.Pipeline)
                    Gst.parse_launch(
                        "filesrc name=src ! " +
                        "decodebin        ! " +
                        "autoaudiosink");
            } catch (GLib.Error e) {
                GLib.warning("There was an error creating the pipeline");
            }

            var bus = pipe.get_bus();
            bus.add_signal_watch();
            bus.message.connect(message_received);

            var src = pipe.get_by_name("src");
            src.set_property("location", path);

            pipe.set_state(Gst.State.PAUSED);
        }

        /**
         * Returns the completion fraction of the media being played.
         * @param duration the current calculated duration.
         * @param position the current position.
         * @return the completion fraction of the media being played.
         */
        public double get_completion(out int64 duration = null,
                                     out int64 position = null) {
            duration = -1;
            position = -1;

            if (!pipe.query_duration(Gst.Format.TIME, out duration))
                return -1;
            if (!pipe.query_position(Gst.Format.TIME, out position))
                return -1;

            return (double)position/duration;
        }

        /* Nessage received handler. */
        private void message_received(Gst.Message message) {
            if (message.type ==Gst.MessageType.EOS) {
                pipe.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, 0);
                pipe.set_state(Gst.State.PAUSED);
            }
        }

        /**
         * Starts playing.
         */
        public void play() {
            pipe.set_state(Gst.State.PLAYING);
        }

        /**
         * Stops playing.
         */
        public void stop() {
            pipe.set_state(Gst.State.NULL);
        }

        /**
         * Seeks a fraction while playing.
         */
        public bool seek(double percentage) {
            int64 duration = -1;
            if (!pipe.query_duration(Gst.Format.TIME, out duration))
                return false;
            if (duration == 0)
                return false;
            int64 nsecs = (int64)(duration * percentage);
            pipe.seek_simple(Gst.Format.TIME,
                             Gst.SeekFlags.ACCURATE|Gst.SeekFlags.FLUSH,
                             nsecs);
            return true;
        }
    }
}
