/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

namespace Rolas {

    /**
     * Class for rola views.
     */
    [GtkTemplate (ui = "/mx/unam/ciencias/Rolas/rola-view.ui")]
    public class RolaView : Gtk.Box {

        /* The default cover. */
        private const string DEFAULT_COVER =
            "media-optical-cd-audio-symbolic";

        /* The cache. */
        private static Database.Cache cache;
        /* Image data memoization. */
        private static Gee.TreeMap<string, GLib.Bytes> all_data;
        /* Pixbufs memoization. */
        private static Gee.TreeMap<string, Gdk.Pixbuf> all_covers;
        /* The default cover pixbuf. */
        private static Gdk.Pixbuf default_cover_pixbuf;
        /* The width. */
        private const int WIDTH = 200;
        /* The height. */
        private const int HEIGHT = 200;

        /* The cover image. */
        [GtkChild]
        private Gtk.Image cover_image;

        /* The title label. */
        [GtkChild]
        private Gtk.Label title_label;

        /* The artist label. */
        [GtkChild]
        private Gtk.Label artist_label;

        /**
         * Creates a new RolaView with the given rola.
         */
        public RolaView(Database.Rola rola) {
            cover_image.pixbuf = get_cover(rola);
            title_label.label = rola.title;
            artist_label.label = rola.artist;
        }

        /**
         * Returns the default cover.
         * @return the default cover.
         */
        public static Gdk.Pixbuf default_cover() {
            if (default_cover_pixbuf == null) {
                Gtk.IconTheme icon_theme = Gtk.IconTheme.get_default();
                try {
                    var p = icon_theme.load_icon(DEFAULT_COVER, 200, 0);
                    default_cover_pixbuf = p;
                } catch (GLib.Error error) {
                    GLib.warning(@"Error loading icon: $(error.message))");
                }
            }
            return default_cover_pixbuf;
        }

        /* Gets the cover data from cache. */
        private static GLib.Bytes? get_cover_data_cache(Database.Rola rola) {
            var c = get_cache();
            if (!c.exists(rola.path))
                return null;
            var data = c.get_cache_data(rola.path);
            return new GLib.Bytes(data);
        }

        /* Gets the cover data from ID3. */
        private static GLib.Bytes get_cover_data_tags(Database.Rola rola) {
            var tags = new Database.FileTags(rola.path);
            var cd = tags.cover_picture;
            get_cache().set_cache_data(rola.path, cd);
            tags = null;
            return new GLib.Bytes(cd);
        }

        /* Gets the cover data. */
        private static GLib.Bytes get_cover_data(Database.Rola rola) {
            var cd = get_cover_data_cache(rola);
            if (cd == null)
                return get_cover_data_tags(rola);
            return cd;
        }

        /**
         * Returns the cover for a rola.
         * @param rola the rola.
         * @return the cover for a rola.
         */
        public static Gdk.Pixbuf get_cover(Database.Rola rola) {
            var covers = get_cached_covers();
            if (covers.has_key(rola.path))
                return covers[rola.path];
            var data = get_cached_data();
            if (data.has_key(rola.path)) {
                covers[rola.path] = create_cover(data[rola.path]);
                return covers[rola.path];
            }
            data[rola.path] = get_cover_data(rola);
            covers[rola.path] = create_cover(data[rola.path]);
            return covers[rola.path];
        }

        /* Creates a cover from data. */
        private static Gdk.Pixbuf create_cover(GLib.Bytes data) {
            var mis = new GLib.MemoryInputStream.from_data(data.get_data(),
                                                           null);
            try {
                var pixbuf = new Gdk.Pixbuf.from_stream(mis);
                return pixbuf.scale_simple(WIDTH, HEIGHT,
                                           Gdk.InterpType.BILINEAR);
            } catch (GLib.Error e) {
                return default_cover();
            }
        }

        /* Get the cached covers. */
        private static Gee.TreeMap<string, Gdk.Pixbuf> get_cached_covers() {
            if (all_covers == null)
                all_covers = new Gee.TreeMap<string, Gdk.Pixbuf>();
            return all_covers;
        }

        /* Get the cached data. */
        private static Gee.TreeMap<string, GLib.Bytes> get_cached_data() {
            if (all_data == null)
                all_data = new Gee.TreeMap<string, GLib.Bytes>();
            return all_data;
        }

        /* Gets the cache. */
        private static Database.Cache get_cache() {
            if (cache == null)
                cache = new Database.Cache();
            return cache;
        }
    }
}
