/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Rolas.Test {

    /**
     * Test cases for {@link Compiler.Analyzer}.
     */
    public class TestAnalyzer : Test.TestCase {

       
        /**
         * Constructs the test suite.
         */
        public TestAnalyzer() {
            base("TestAnalyzer");
            add_test("test_analyze_matches",     test_analyze_matches);
            add_test("test_analyze_parenthesis", test_analyze_parenthesis);
            add_test("test_analyze_unary_op",    test_analyze_unary_op);
            add_test("test_analyze_binary_ops",  test_analyze_binary_ops);
            add_test("test_analyze_errors",      test_analyze_errors);
        }

        /**
         * Test for analyze matches with {@link Analyzer}.
         */
        public void test_analyze_matches() {
            Compiler.Tree tree = null;
            var tokenizer = new Compiler.Tokenizer();
            var analyzer = new Compiler.Analyzer();
            try {
                var ts = tokenizer.tokenize("title:\"title x\"");
                tree = analyzer.analyze(ts);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(false);
            }
            GLib.assert(tree.token.kind ==
                        Compiler.Token.Kind.COLON);
            GLib.assert(tree.left.token.kind ==
                        Compiler.Token.Kind.FIELD);
            GLib.assert(tree.right.token.kind ==
                        Compiler.Token.Kind.STRING);
            GLib.assert(tree.left.token.original == "title");
            GLib.assert(tree.right.token.original == "title x");
        }

        /**
         * Test for analyze parenthesis with {@link Analyzer}.
         */
        public void test_analyze_parenthesis() {
            Compiler.Tree tree = null;
            var tokenizer = new Compiler.Tokenizer();
            var analyzer = new Compiler.Analyzer();

            var input = "title:\"title x\"";
            for (int i = 0; i < 5; i++) {
                try {
                    var ts = tokenizer.tokenize(input);
                    tree = analyzer.analyze(ts);
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                }
                GLib.assert(tree.token.kind ==
                            Compiler.Token.Kind.COLON);
                GLib.assert(tree.left.token.kind ==
                            Compiler.Token.Kind.FIELD);
                GLib.assert(tree.right.token.kind ==
                            Compiler.Token.Kind.STRING);
                GLib.assert(tree.left.token.original == "title");
                GLib.assert(tree.right.token.original == "title x");
                input = "(" + input + ")";
            }
        }

        /**
         * Test for analyze unary op with {@link Analyzer}.
         */
        public void test_analyze_unary_op() {
            Compiler.Tree tree = null;
            var tokenizer = new Compiler.Tokenizer();
            var analyzer = new Compiler.Analyzer();

            var input = "title:\"title x\"";
            for (int i = 0; i < 5; i++) {
                try {
                    var ts = tokenizer.tokenize(input);
                    tree = analyzer.analyze(ts);
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                }
                var t = tree;
                for (int j = 0; j < i; j++) {
                    GLib.assert(t.token.kind ==
                                Compiler.Token.Kind.UNARY_OPERATOR);
                    GLib.assert(t.token.original == "!");
                    t = t.right;
                }
                GLib.assert(t.token.kind == Compiler.Token.Kind.COLON);
                GLib.assert(t.left.token.kind ==
                            Compiler.Token.Kind.FIELD);
                GLib.assert(t.right.token.kind ==
                            Compiler.Token.Kind.STRING);
                GLib.assert(t.left.token.original == "title");
                GLib.assert(t.right.token.original == "title x");
                input = "!(" + input + ")";
            }
        }

        /**
         * Test for analyze binary ops with {@link Analyzer}.
         */
        public void test_analyze_binary_ops() {
            Compiler.Tree tree = null;
            var tokenizer = new Compiler.Tokenizer();
            var analyzer = new Compiler.Analyzer();

            var input = "title:\"title x\"";
            for (int i = 0; i < 5; i++) {
                try {
                    var ts = tokenizer.tokenize(input);
                    tree = analyzer.analyze(ts);
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                }
                var t = tree;
                for (int j = 0; j < i; j++) {
                    GLib.assert(t.token.kind ==
                                Compiler.Token.Kind.BINARY_OPERATOR);
                    GLib.assert(t.token.original == "&");
                    t = t.left;
                }
                GLib.assert(t.token.kind == Compiler.Token.Kind.COLON);
                GLib.assert(t.left.token.kind ==
                            Compiler.Token.Kind.FIELD);
                GLib.assert(t.right.token.kind ==
                            Compiler.Token.Kind.STRING);
                GLib.assert(t.left.token.original == "title");
                GLib.assert(t.right.token.original == "title x");
                input = "title:\"title x\"&" + input;
            }
        }

        /**
         * Test for analyze errors with {@link Analyzer}.
         */
        public void test_analyze_errors() {
            var tokenizer = new Compiler.Tokenizer();
            var analyzer = new Compiler.Analyzer();
            try {
                var ts = tokenizer.tokenize(":\"title x\"");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected colon");
            }
            try {
                var ts = tokenizer.tokenize("title:");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Missing string");
            }
            try {
                var ts = tokenizer.tokenize("t:z t:y");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected token 'title'");
            }
            try {
                var ts = tokenizer.tokenize("t:z&");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Incomplete expresion");
            }
            try {
                var ts = tokenizer.tokenize("t:z!");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected token '!'");
            }
            try {
                var ts = tokenizer.tokenize("(t:z)(t:a)");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected token '('");
            }
            try {
                var ts = tokenizer.tokenize("t:z)");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unbalanced parenthesis");
            }
            try {
                var ts = tokenizer.tokenize("(t:z");
                analyzer.analyze(ts);
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Missing closing parenthesis");
            }
        }
    }
}
