/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Rolas.Test {

    /**
     * Test cases for {@link Compiler.Tokenizer}.
     */
    public class TestTokenizer : Test.TestCase {

        /* Test strings. */
        private const string[] FIELDS = { "title", "artist", "album",
                                          "tit", "art", "alb",
                                          "t", "a", "l" };
        private const string[] ALPHAS = { "alpha", "beta", "gamma" };
        private const string[] ALPHA_NUMS = { "alpha1", "be2ta", "g3amma" }; 
        private const string[] OPERATORS = { "(", ")", "&", "|", "!" };
       
        /**
         * Constructs the test suite.
         */
        public TestTokenizer() {
            base("TestTokenizer");
            add_test("test_tokenize_fields",    test_tokenize_fields);
            add_test("test_tokenize_alphas",    test_tokenize_alphas);
            add_test("test_tokenize_apha_nums", test_tokenize_alpha_nums);
            add_test("test_tokenize_operators", test_tokenize_operators);
            add_test("test_tokenize_random",    test_tokenize_random);
        }

        /**
         * Test for tokenize fields with {@link Tokenizer}.
         */
        public void test_tokenize_fields() {
            var tokenizer = new Compiler.Tokenizer();
            for (int i = 0; i < FIELDS.length; i++) {
                Gee.ArrayList<Compiler.Token> tokens;
                try {
                    tokens = tokenizer.tokenize(FIELDS[i]);
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                    continue;
                }
                GLib.assert(tokens.size == 1);
                foreach (var token in tokens) {
                    GLib.assert(token.kind ==
                                Compiler.Token.Kind.FIELD);
                }
            }
        }

        /**
         * Test for tokenize alphas with {@link Tokenizer}.
         */
        public void test_tokenize_alphas() {
            var tokenizer = new Compiler.Tokenizer();
            for (int i = 0; i < ALPHAS.length; i++) {
                Gee.ArrayList<Compiler.Token> tokens;
                try {
                    tokens = tokenizer.tokenize(ALPHAS[i]);
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                    continue;
                }
                GLib.assert(tokens.size == 1);
                foreach (var token in tokens) {
                    GLib.assert(token.kind ==
                                Compiler.Token.Kind.STRING);
                }
            }
        }

        /**
         * Test for tokenize alphanumericals with {@link Tokenizer}.
         */
        public void test_tokenize_alpha_nums() {
            var tokenizer = new Compiler.Tokenizer();
            for (int i = 0; i < ALPHA_NUMS.length; i++) {
                Gee.ArrayList<Compiler.Token> tokens;
                try {
                    tokens = tokenizer.tokenize(ALPHA_NUMS[i]);
                    GLib.assert(false);
                } catch (Compiler.SyntaxError error) {
                    continue;
                }
            }
            for (int i = 0; i < ALPHA_NUMS.length; i++) {
                Gee.ArrayList<Compiler.Token> tokens;
                try {
                    tokens = tokenizer.tokenize("\"" + ALPHA_NUMS[i] + "\"");
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                    continue;
                }
                GLib.assert(tokens.size == 1);
                foreach (var token in tokens) {
                    GLib.assert(token.kind ==
                                Compiler.Token.Kind.STRING);
                    GLib.assert(token.original == ALPHA_NUMS[i]);
                }
            }
        }

        /**
         * Test for tokenize operators with {@link Tokenizer}.
         */
        public void test_tokenize_operators() {
            var tokenizer = new Compiler.Tokenizer();
            for (int i = 0; i < OPERATORS.length; i++) {
                Gee.ArrayList<Compiler.Token> tokens;
                try {
                    tokens = tokenizer.tokenize(OPERATORS[i]);
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                    continue;
                }
                GLib.assert(tokens.size == 1);
                if (i == 0) {
                    foreach (var token in tokens) {
                        GLib.assert(token.kind ==
                                    Compiler.Token.Kind.PAR_LEFT);
                    }
                } else if (i == 1) {
                    foreach (var token in tokens) {
                        GLib.assert(token.kind ==
                                    Compiler.Token.Kind.PAR_RIGHT);
                    }
                } else if (i < 4) {
                    foreach (var token in tokens) {
                        GLib.assert(token.kind ==
                                    Compiler.Token.Kind.BINARY_OPERATOR);
                    }
                } else {
                    foreach (var token in tokens) {
                        GLib.assert(token.kind ==
                                    Compiler.Token.Kind.UNARY_OPERATOR);
                    }
                }
            }
        }

        /* Random entry in a string array. */
        private string get_random(string[] a, GLib.Rand rand) {
            return a[rand.int_range(0, a.length)];
        }

        /**
         * Test for tokenize a random input with {@link Tokenizer}.
         */
        public void test_tokenize_random() {
            var rand = new GLib.Rand();
            int n = rand.int_range(1, 10);
            string[] spaces = { " ", "  ", "   ", "    " };
            string str1 = "";
            string str2 = "";
            for (int i = 0; i < n; i++) {
                str2 += get_random(spaces, rand);
                string s = get_random(FIELDS, rand);
                str1 += s;
                str2 += s;
                str2 += get_random(spaces, rand);

                str2 += get_random(spaces, rand);
                s = get_random(OPERATORS, rand);
                str1 += s;
                str2 += s;
                str2 += get_random(spaces, rand);

                str2 += get_random(spaces, rand);
                s = get_random(ALPHAS, rand);
                str1 += s;
                str2 += s;
                str2 += get_random(spaces, rand);

                str1 += get_random(spaces, rand);
            }
            var tokenizer = new Compiler.Tokenizer();
            Gee.ArrayList<Compiler.Token> tokens1, tokens2;
            try {
                tokens1 = tokenizer.tokenize(str1);
                tokens2 = tokenizer.tokenize(str2);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(false);
                return;
            }
            GLib.assert(tokens1.size == tokens2.size);
            Gee.Iterator<Compiler.Token> it1, it2;
            it1 = tokens1.iterator();
            it2 = tokens2.iterator();
            while (it1.has_next() && it2.has_next()) {
                GLib.assert(it1.next() && it2.next());
                var token1 = it1.get();
                var token2 = it2.get();
                GLib.assert(token1.kind == token2.kind);
                GLib.assert(token1.original == token2.original);
            }
        }
    }
}
