/*
 * This file is part of rolas.
 *
 * Copyright © 2018 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Rolas.Test {

    /**
     * Test cases for {@link Compiler.Translator}.
     */
    public class TestTranslator : Test.TestCase {

        private const string TITLE_X = """
SELECT rolas.path
FROM rolas
JOIN performers ON rolas.id_performer = performers.id_performer
JOIN albums ON rolas.id_album = albums.id_album
WHERE  ( rolas.title  LIKE "%title x%" ) 
""";

        private const string NOT_TITLE_X = """
SELECT rolas.path
FROM rolas
JOIN performers ON rolas.id_performer = performers.id_performer
JOIN albums ON rolas.id_album = albums.id_album
WHERE  (  NOT  (  ( rolas.title  LIKE "%title x%" )  ) ) 
""";

        private const string AND_X = """
SELECT rolas.path
FROM rolas
JOIN performers ON rolas.id_performer = performers.id_performer
JOIN albums ON rolas.id_album = albums.id_album
WHERE  ( ( rolas.title  LIKE "%tx%" )  AND  ( performers.name  LIKE "%ax%" )  ) 
""";

        /**
         * Constructs the test suite.
         */
        public TestTranslator() {
            base("TestTranslator");
            add_test("test_translate_matches",     test_translate_matches);
            add_test("test_translate_parenthesis", test_translate_parenthesis);
            add_test("test_translate_unary_op",    test_translate_unary_op);
            add_test("test_translate_binary_ops",  test_translate_binary_ops);
            add_test("test_translate_errors",      test_translate_errors);
        }

        /**
         * Test for translate matches with {@link Translator}.
         */
        public void test_translate_matches() {
            var translator = new Compiler.Translator();
            string output = null;
            try {
                output = translator.translate("title:\"title x\"");
            } catch (Compiler.SyntaxError error) {
                GLib.assert(false);
            }
            GLib.assert(output == TITLE_X);
        }

        /**
         * Test for translate parenthesis with {@link Translator}.
         */
        public void test_translate_parenthesis() {
            var translator = new Compiler.Translator();
            var input = "title:\"title x\"";
            for (int i = 0; i < 5; i++) {
                string output = null;
                try {
                    output = translator.translate(input);
                } catch (Compiler.SyntaxError error) {
                    GLib.assert(false);
                }
                GLib.assert(output == TITLE_X);
                input = "(" + input + ")";
            }
        }

        /**
         * Test for translate unary op with {@link Translator}.
         */
        public void test_translate_unary_op() {
            var translator = new Compiler.Translator();
            var input = "!(title:\"title x\")";
            string output = null;
            try {
                output = translator.translate(input);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(false);
            }
            GLib.assert(output == NOT_TITLE_X);
        }

        /**
         * Test for translate binary ops with {@link Translator}.
         */
        public void test_translate_binary_ops() {
            var translator = new Compiler.Translator();
            var input = "(t:\"tx\") & (a:\"ax\")";
            string output = null;
            try {
                output = translator.translate(input);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(false);
            }
            stderr.printf("%s\n", output);
            GLib.assert(output == AND_X);
        }

        /**
         * Test for translate errors with {@link Translator}.
         */
        public void test_translate_errors() {
            var translator = new Compiler.Translator();
            try {
                translator.translate(":\"title x\"");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected colon");
            }
            try {
                translator.translate("title:");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Missing string");
            }
            try {
                translator.translate("t:z t:y");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected token 'title'");
            }
            try {
                translator.translate("t:z&");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Incomplete expresion");
            }
            try {
                translator.translate("t:z!");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected token '!'");
            }
            try {
                translator.translate("(t:z)(t:a)");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unexpected token '('");
            }
            try {
                translator.translate("t:z)");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Unbalanced parenthesis");
            }
            try {
                translator.translate("(t:z");
                GLib.assert(false);
            } catch (Compiler.SyntaxError error) {
                GLib.assert(error.message == "Missing closing parenthesis");
            }
        }
    }
}
